package spiropro;



import java.io.DataOutputStream;
import java.lang.Integer;
import java.io.DataInputStream;
import java.io.FileOutputStream;
import java.io.File;
import java.nio.DoubleBuffer;
import javax.microedition.io.StreamConnection;
import javax.microedition.io.Connector;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

import java.io.IOException;

public class SpiroProApp implements Runnable{

	String serverURL;
	private Thread spiroThread = null;
	String user;
	

	public SpiroProApp(String serverURLarg,String username) {	
		user=username;
		serverURL = serverURLarg;

		//construct the local thread
		if (spiroThread == null) {
			spiroThread = new Thread(this,"SpiroThread");
			spiroThread.start();
		}

	}

	//this is the code that will run by the thread
	public void run() {


		launchProgress();
		createConnection();

		stop();
	}


	//terminate current thread
	public void stop() {

		spiroThread = null;

		if (spiroThread != null) {
			spiroThread.interrupt();
		}

	}


	
	SpiroTest spiroObj;
	
	//float sampleStep = 1.339f/379f;
	float sampleStep = 12.125f/3589f;

	// sampling frequency
	int fs = 250;

	//Tsampling = 1/250;
	float Ts = 1f/fs;

	File spiroFile;
	FileOutputStream fos;
	SpiroTestProgress spiroProgress;
	static boolean progressIsOpen = false;
	static StreamConnection sc;
	static DataOutputStream dos;	
	static DataInputStream dis;
	byte[] initCommand1 = {0x20,0x20};
	byte[] initCommand2 = {0x51,0x51};
	byte[] initCommand3= {0x17,0x17};

	byte[] initResponse1 = new byte[30];
	byte[] deviceID = new byte[6];
	byte[] ack=new byte[5];
	static String deviceIDString;
	byte[] initResponse2 = new byte[7];
	byte[] initResponse3 = new byte[9];

	byte[] startCommand = {0x40,0x40};
	byte[] startResponse = new byte[5];
	byte[] startSuccessResponse = {-4,0x40,0x01,-3,0x40};	

	byte[] stopCommand = {0x41,0x41};
	byte[] stopResponse = new byte[3];

	
	 //* a simple method, just to display the progress of the spiro test
	 
	public void launchProgress() {
		spiroProgress = new SpiroTestProgress();
		spiroProgress.setVisible(true);
	}

	public void createConnection() {	

		try {
			sc = (StreamConnection)Connector.open("btspp://" +serverURL +":1",3);

			 dos = new DataOutputStream(sc.openOutputStream());	
			 dis = new DataInputStream(sc.openDataInputStream());

			//once we re connected, change the status of the progress bar and the label
			spiroProgress.getJProgressBar().setIndeterminate(false);
			spiroProgress.jLabel.setText("");

			//start the spiro test
			spiroTest();
		}
		catch (IOException e) {				
			try{Thread.sleep(1000);}catch(InterruptedException e1) {}
			if (progressIsOpen) {
				createConnection();
			}
			else {
				stop();
			}
		}catch(Exception ex){
			spiroProgress.dispose();
			progressIsOpen = false;
			JOptionPane.showMessageDialog(new JFrame(),"   .\n  ->.","",JOptionPane.WARNING_MESSAGE);
			stop();
		}
	}

	public void spiroTest() {

		try {

			

			dos.write(initCommand1);
			dos.flush();
			dis.readFully(initResponse1);



			dos.write(initCommand2);
			dos.flush();
			dis.readFully(initResponse2);

			dos.write(initCommand3);
			dos.flush();
			dis.readFully(initResponse3);

			dos.write(startCommand);
			dos.flush();
			dis.readFully(startResponse);
			String xx="";
			for(int i=0;i<startResponse.length;i++){
				xx=xx+" "+Integer.toHexString(startResponse[i]);
			}

			spiroProgress.jLabel.setText("   ");
			spiroProgress.getJProgressBar().setIndeterminate(false);

			int numberOfSamples = 0;

			int maxSamples = 30000*2;
			DoubleBuffer flowBuffer = DoubleBuffer.allocate(maxSamples);
			DoubleBuffer volumeBuffer = DoubleBuffer.allocate(maxSamples);

			float volume = 0;			
			float oldFlow = 0;

			long maxDuration = 20000;
			//Integer lsb,msb;
			int lsb=0,msb=0;
			Short data;
			Float flow;

			//array that is used to buffer the values of the last second of the measurmnent
			//Note that fsampling = 1/Ts
			double[] eotArray = new double[fs];
			boolean EOT = false;
			//init the array
			java.util.Arrays.fill(eotArray, 0);

			//BEEP TONE - The beep tone lasts approx 2 secs. We dont want to start the timer yet. Wait for 2 secs
			try{Thread.sleep(2000);}catch(InterruptedException e1) {}

			long startTime = System.currentTimeMillis();

			//the spirometer test will last 'duration'  msecs
			while (System.currentTimeMillis() - startTime < maxDuration & !EOT) {

				
				try {
	                

	                lsb = (short)dis.read();
	                msb = (short)dis.read();
	                if(lsb<0)
	                    lsb=(short)(lsb+256);
	                if(msb<0)
	                    msb=(short)(msb+256);

	                
	            }
	                catch (Exception ex) {
	                continue;
	            }
	               
	           

	            
	            data=(short)(lsb+256*msb & 0x1FFF);
				
				flow = data*sampleStep;
				flowBuffer.put(flow);

				//trapezoid approximate method to calculate the total volume exhaled by the patient				
				volume += ((flow + oldFlow)/2)*Ts;
				volumeBuffer.put(volume);

				oldFlow = flow;


				
				if (System.currentTimeMillis() - startTime > 5000) {
					for (int i= fs - 1; i>0; i--) {

						eotArray[i] = eotArray[i-1]; 

					}
					eotArray[0] = volume;

								

					if (eotArray[0] - eotArray[eotArray.length-1] > 0.025) {
						EOT = false;
					}
					else {
						EOT = true;
					}										

					if (EOT) {
					}
				}

				numberOfSamples += 1;

			}

			float duration = (System.currentTimeMillis() - startTime)/1000f;

			
			byte[] closecommand={0x52,0x52};
            dos.write(stopCommand);
            dos.flush();
            dis.read(ack);          
            dos.write(closecommand);
            dos.flush();
            dis.read(ack);
			dis.close();
			dos.close();
			sc.close();
			//dispose the progress bar
			spiroProgress.dispose();
			progressIsOpen = false;

			//process spirometry test data
			spiroProcess(flowBuffer,volumeBuffer, numberOfSamples,duration);
		}
		catch (IOException e) {				
		}
	}

//	a function that processes data received by the spirometer and decides for the Start of Test criteria
//	as defined in Standardization of Spirometry
	public void spiroProcess(DoubleBuffer flowBuffer,DoubleBuffer volumeBuffer, int numberOfSamples, float duration) {


		double maxDifference = 0;
		double difference;
		double lamda,b;
		int n = (int)(0.08*fs);
		int index = 0;
		int maxIndex = 0;

		//TIME ZERO CALCULATION - calculated from max slope of Volume-Time curve as defined in Start of Test criteria
		//moving average for 80ms window. That is approx 20 samples. We have to find the maximum slope of the Volume-Timecurve (L/s)
		//Please refer to site for more info http://erj.ersjournals.com/cgi/reprint/26/2/319.pdf
		while (volumeBuffer.position() < volumeBuffer.capacity() - n+1) {
			difference = -volumeBuffer.get() + volumeBuffer.get(index + n-1);

			if (difference > maxDifference ) {
				//the index of the max slope of flow will be obviously be the average of index and n.
				maxDifference = difference;
				maxIndex = index;
			}

			//increment counter and set the new position of the floatbuffer. 
			index++;
			volumeBuffer.position(index);
		}

		//lamda = (y1-y2)/(x1-x2)
		//b = y1 - lamda*x1
		//Time_zero_of_Test = -b/lamda. that is for y=0.
		lamda = (volumeBuffer.get(maxIndex + n-1) - volumeBuffer.get(maxIndex))/(n-1);
		b = volumeBuffer.get(maxIndex) - lamda*maxIndex;
		double timeZero = -b/lamda;

		//Spirometry standard defines that we should include values, at least 0.25s prior to exhalation.
		double offset = 0.25f*fs;
		double xStart;
		if (timeZero > offset) {					
			xStart = timeZero - offset;
		}
		else {
			xStart = 0;
		}


		//the new volume array will contain only the values that we're interested in and will start from the 
		//'time zero' - 0.25s and end to the last sample of the test.		
		double[] volumeArray = new double[numberOfSamples - (int)xStart];
		volumeBuffer.position((int)xStart);
		volumeBuffer.get(volumeArray, 0 , numberOfSamples-(int)xStart);

		//same as the volumeArray
		double[] flowArray = new double[numberOfSamples - (int)xStart];
		flowBuffer.position((int)xStart);
		flowBuffer.get(flowArray, 0, numberOfSamples-(int)xStart);
		

		//EV CALCULATION - the value of volume exhaled at timeZero value
		double EV = volumeArray[(int)(offset)];

		//FVC CALCULATION - total volume of air exhaled by the subject
		double FVC = volumeArray[volumeArray.length-1];

		//FEV1 CALCULATION - volume of air exhaled in first second
		//1 sec is equal to 250 samples
		double FEV1;
		if (volumeArray.length >= (int)((offset) + fs) ) {
			FEV1 = volumeArray[(int)((offset) + fs)];
		}
		else {
			FEV1 = volumeArray[volumeArray.length-1];
		}

		//FEV3 CALCULATION - volume of air exhaled in first three seconds
		//3 sec is equal to 750 samples
		double FEV3;
		if (volumeArray.length >= (int)(offset + 3*fs)) {
			FEV3 = volumeArray[(int)(offset + 3*fs)];
		}
		else {
		FEV3 = volumeArray[volumeArray.length-1];
		}

		//PEF CALCULATION - that is the maximum value of flow (L/s)
		double PEF = 0;
		for (int i=(int)offset;i<flowArray.length;i++) {

			if (flowArray[i] > PEF) {
				PEF = flowArray[i];
			}
		}

		//FEFx indexes CALCULATION - Index of the array which corresponds to x % of the FVC. These indexes will
		//be used to calculate the FEFx % values.
		int FEF25Index = 0,FEF50Index = 0,FEF75Index = 0;
		//FEF25 Index
		for (int i=(int)offset;i<volumeArray.length-1;i++) {
			if (volumeArray[i] <= 0.25f*FVC & volumeArray[i+1] >= 0.25f*FVC) {
				FEF25Index = i;
				break;
			}
		}
		

		//FEF50 Index		
		for (int i=(int)offset;i<volumeArray.length-1;i++) {
			if (volumeArray[i] <= 0.5f*FVC & volumeArray[i+1] >= 0.5f*FVC) {
				FEF50Index = i;
				break;
			}
		}

		//FEF75 Index
		for (int i=(int)offset;i<volumeArray.length-1;i++) {
			if (volumeArray[i] <= 0.75f*FVC & volumeArray[i+1] >= 0.75f*FVC) {
				FEF75Index = i;
				break;
			}
		}

		//FEF25 CALCULATION - Average flow exhaled during first 25% of test
		int k = (int)offset;
		double sum = 0;
		for (k=(int)offset;k<=FEF25Index;k++) {
			sum += flowArray[k];
		}
		double FEF25 = sum/k;

		//FEF50 CALCULATION - Average flow exhaled during first 50% of test
		sum = 0;
		for (k=(int)offset;k<=FEF50Index;k++) {
			sum += flowArray[k];
		}
		double FEF50 = sum/k;

		//FEF75 CALCULATION - Average flow exhaled during first 75% of test
		sum = 0;
		for (k=(int)offset;k<=FEF75Index;k++) {
			sum += flowArray[k];
		}
		double FEF75 = sum/k;

		//FEF25-75 CALCULATION - Average flow exhaled during 25% to 75% of test
		sum = 0;
		for (int i=FEF25Index; i<=FEF75Index;i++) {
			sum += flowArray[i];
		}
		double FEF2575 = sum/(FEF75Index - FEF25Index);



		//START OF TEST CRITERIA
		if (EV > 0.05f*FVC) {
			System.out.println("Start of Test criteria not met. EV = " +EV +"> 0.05*FVC=" +FVC);
		}
		else if (EV > 0.15f) {
			System.out.println("Start of Test criteria not met. EV = " +EV +"> 0.15L");

		}
	
		
		
		//now create and construct the Spiro Test object
		spiroObj = new SpiroTest();
		
		spiroObj.setFileName(DateTimeTools.getDateTimeString2());
		spiroObj.setFilePath(Consts.SPO2_FOLDER +"\\Spiro_" +DateTimeTools.getDateTimeString2() +".spi");
		spiroObj.setDeviceID(deviceIDString);
		spiroObj.setDateTime(DateTimeTools.getDateTimeString2());
		spiroObj.setSampleRate(String.valueOf(fs));
		spiroObj.setFVC(String.valueOf(FVC));
		spiroObj.setFEV1(String.valueOf(FEV1));
		spiroObj.setFEV3(String.valueOf(FEV3));
		spiroObj.setFEV1toFVC(String.valueOf((FEV1/FVC)*100f));
		spiroObj.setFEV3toFVC(String.valueOf((FEV3/FVC)*100f));
		spiroObj.setPEF(String.valueOf(PEF));
		spiroObj.setFEF25(String.valueOf(FEF25));
		spiroObj.setFEF50(String.valueOf(FEF50));
		spiroObj.setFEF75(String.valueOf(FEF75));
		spiroObj.setFEF2575(String.valueOf(FEF2575));		
		spiroObj.setVolumeVSTimeData(volumeArray);
		spiroObj.setFLowVSTimeData(flowArray);
		
		//write to file all results
		SpiroTestFileWriter.writeToFile(spiroObj);
		
		
		savelocal sl=new savelocal();
		sl.savelocalspiro(pre(FVC), pre(FEV1),pre( FEV3), pre(((FEV1/FVC)*100f)), pre((FEV3/FVC)*100f), pre(PEF), pre(FEF25), pre(FEF50), pre(FEF75), pre(FEF2575), DateTimeTools.getDateTimeString1(), "Spiro_" +DateTimeTools.getDateTimeString2() +".spi");
		
		//Display the results to monitor
		//display the spiro test using SpiroProFileReader class
		SpiroProFileReader.event=sl.getEventId();
		SpiroProFileReader.processFile(spiroObj.getFilePath(),user);

	}
	
	double pre(double spirod) {
    	
    		return Math.round(spirod*100)/100D;
		}


	public static void main(String[] args) {


	}

}